/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: draw.c,v 1.65 1997/12/24 06:41:38 yuo Exp $
 */

#include "mgp.h"

/* image/ *.[ch] */
extern void bestVisual __P((Display *, int, Image *, Visual **, unsigned int *));

static u_int special;
static u_int align;
static u_int again;
static u_int pause_on;
static u_int pause_line;
static XFontStruct *xfontstruct;

static short ctable[128] = {
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121, 0x2121,
	0x2121, 0x212a, 0x216d, 0x2174, 0x2170, 0x2173, 0x2175, 0x216c,
	0x214a, 0x214b, 0x2176, 0x215c, 0x2124, 0x215d, 0x2125, 0x213f,
	0x2330, 0x2331, 0x2332, 0x2333, 0x2334, 0x2335, 0x2336, 0x2337,
	0x2338, 0x2339, 0x2127, 0x2128, 0x2152, 0x2161, 0x2153, 0x2129,
	0x2177, 0x2341, 0x2342, 0x2343, 0x2344, 0x2345, 0x2346, 0x2347,
	0x2348, 0x2349, 0x234a, 0x234b, 0x234c, 0x234d, 0x234e, 0x234f,
	0x2350, 0x2351, 0x2352, 0x2353, 0x2354, 0x2355, 0x2356, 0x2357,
	0x2358, 0x2359, 0x235a, 0x214e, 0x2140, 0x214f, 0x2130, 0x2132,
	0x212e, 0x2361, 0x2362, 0x2363, 0x2364, 0x2365, 0x2366, 0x2367,
	0x2368, 0x2369, 0x236a, 0x236b, 0x236c, 0x236d, 0x236e, 0x236f,
	0x2370, 0x2371, 0x2372, 0x2373, 0x2374, 0x2375, 0x2376, 0x2377,
	0x2378, 0x2379, 0x237a, 0x2150, 0x2143, 0x2151, 0x2141, 0x2121,
};

static u_int pos_x;
static u_int pos_y;
static u_int again_pos_x;
static u_int last_pos_x;
static u_int last_pos_y;
static u_int string_width;
static u_int cont;
static u_int line;
static u_int no_bg_image;
static u_char xfont[BUFSIZ] = DEFAULT_X_FONT;
static u_char *prefix = NULL, *oldprefix = NULL;
static int prev_tabp = 0;
static struct ctrl *bg_image = NULL;
static struct ctrl *bg_image_prev = NULL;

static void draw_textline_main __P((u_char *, u_int));
static void draw_textline __P((u_char *, u_int));
#ifdef VFLIB
static void draw_vfstring __P((u_char *, int, u_int));
static u_int draw_onechar __P((u_int, u_int, u_int, u_int, u_int, u_int));
#else
static void draw_xstring __P((u_char *, int));
#endif
static void process_direc __P((struct ctrl *));
static void back_gradation __P((struct ctrl_grad *));

static void image_load __P((char *, int, int, int, int, int));
static u_int draw_onechar_x __P((u_char *, u_int, u_int));
static void lcutin __P((u_char *, u_int));
static void rcutin __P((u_char *, u_int));
static void shrink __P((u_char *, u_int));
static int set_position __P((void));
static void timebar __P((u_int));

/* for gradation image */
struct g_color  colors[G_MAXCOLS];
struct g_color  *g_curcol, *g_nextcol;

#ifndef HAVE_USLEEP
#include <sys/time.h>
void
usleep(usec)
	u_int usec;
{
#ifdef HAVE_POLL
	(void)poll(NULL, 0, usec/1000);
#else /* HAVE_POLL */
	struct timeval tv;
	tv.tv_sec  = usec / 1000000;
	tv.tv_usec = usec % 1000000;
	(void)select(0, NULL, NULL, NULL, &tv);
#endif /* HAVE_POLL */
}
#endif /* HAVE_USLEEP */

void
draw_page(page)
	u_int page;
{
	u_int	start_line;
	u_int	tab_depth;
	u_int	end_line;
	u_char	*p;
	struct	ctrl *cp;
	u_int	save_pl_mode = pl_mode;

	if (pause_flag == 0 || cur_page != page) {
		prefix = NULL; oldprefix = NULL; prev_tabp = 0;
		cur_page = page;
		last_pos_x = last_pos_y = pos_x = pos_y = 0;
		start_line = pause_line = 0;
	} else {
		start_line = pause_line;
	}
	pause_flag = 0; pause_on = 0;
	end_line = page_attribute[page].pg_linenum;
	if (pl_mode)
		pl_off();

	for (line = start_line; line <= end_line; line++) {

		if (prev_tabp) {
			prefix = oldprefix;
			oldprefix = NULL;
			prev_tabp = 0;
		}
		/* process default control */
		if (!(page_attribute[page].pg_flag & PGFLAG_NODEF))
			for (cp = default_control[line]; cp; cp = cp->ct_next)
				process_direc(cp);

		/* process normal control */
		for (cp = page_control[page][line]; cp; cp = cp->ct_next)
			process_direc(cp);

		/* tab control */
		p = page_data[page][line];
		if (p && *p == '\t') {
			if (! prev_tabp) {
				oldprefix = prefix;
				prev_tabp = 1;
			}
			tab_depth = 0;
			p++;
			while (*p == '\t') {
				tab_depth++;
				p++;
			}
			for (cp = tab_control[tab_depth]; cp; cp = cp->ct_next)
				process_direc(cp);
		}

		if (line == 0) {
			no_bg_image = 0;
			if (bg_image && bg_image_prev
			 && ctlcmp(bg_image, bg_image_prev) == 0) {
				; /* bimage or bgrad, same as before */
			} else if (bg_image && bg_image->ct_op == CTL_BIMAGE) {
				/* bimage, differs */
				image_load(bg_image->ctm_fname,
					bg_image->ctm_numcolor,
					bg_image->ctm_ximagesize,
					bg_image->ctm_yimagesize, 1,
					bg_image->ctm_zoomflag);
			} else if (bg_image && bg_image->ct_op == CTL_BGRAD) {
				/* bgrad, differs */
				back_gradation(&bg_image->ct_val.ctrl_grad);
			} else {
				/* plain, regardless same or differs */
				XSetWindowBackground(display, window,
					back_color);
				no_bg_image = 1;
			}
			XClearWindow(display, window);
			bg_image_prev = bg_image;
			bg_image = NULL;
			if (t_fin)
				timebar(page);
		}

		if (pause_flag)
			break;

		if (p)
			draw_textline_main(p, page);

		pos_y += VERT_STEP(char_size);
	}
	if (save_pl_mode)
		pl_on();
	if (pg_mode)
		pg_on(page, (line > end_line));
}

static void
draw_textline_main(data, page)
	u_char *data;
	u_int page;
{
	switch(special) {
	case SP_SHRINK:
		shrink(data, page);
		break;

	case SP_LCUTIN:
		lcutin(data, page);
		break;

	case SP_RCUTIN:
		rcutin(data, page);
		break;

	default:
		break;
	}
	draw_textline(data, page);
	special = 0;
}

static void
draw_textline(data, page)
	u_char *data;
	u_int page;
{

	if (no_bg_image) {
		XFillRectangle(display, pixmap, gcback, 0, 0,
			window_width, VERT_STEP(char_size));
	} else {
		XFillRectangle(display, pixmap, gcnone, 0, 0,
			window_width, VERT_STEP(char_size));
		if (fore_color != depth_mask)
			XFillRectangle(display, maskpix, gcnone, 0, 0,
				window_width, VERT_STEP(char_size));
	}

	if (prefix && *prefix) {
#ifdef VFLIB
		draw_vfstring(prefix, 1, page);
		draw_vfstring(data, 0, page);
#else
		draw_xstring(prefix, 1);
		draw_xstring(data, 0);
#endif /*VFLIB*/
	} else {
#ifdef VFLIB
		draw_vfstring(data, 1, page);
#else
		draw_xstring(data, 1);
#endif
	}

	pos_x = set_position();

	if (cont) {
		pos_x = last_pos_x;
		pos_y = last_pos_y;
		cont = 0;
	}
	if (again) {
		pos_x = again_pos_x;
		pos_y = last_pos_y;
		again = 0;
	}
	if (no_bg_image) {
		XCopyArea(display, pixmap, window, gcall, 0, 0,
				string_width, VERT_STEP(char_size),
				pos_x, VERT_POS(pos_y, char_size));
	} else {
		if (fore_color != depth_mask)
			XCopyArea(display, maskpix, window, gcmask, 0, 0,
				string_width, VERT_STEP(char_size),
				pos_x, VERT_POS(pos_y, char_size));
		XCopyArea(display, pixmap, window, gcor, 0, 0,
				string_width, VERT_STEP(char_size),
				pos_x, VERT_POS(pos_y, char_size));
	}

	last_pos_x = pos_x + string_width;
	again_pos_x = pos_x;
	last_pos_y = pos_y;
}

#ifdef VFLIB
static void
draw_vfstring(data, init, page)
	u_char *data;
	int init;
	u_int page;
{
	u_char *p;
	u_int kanji = 0;
	u_int code;
	u_int char_len;

	p = data;
	if (init) string_width = 0;

	while (*p != '\n' && *p != 0) {
		if (*p == 0x1b && *(p+1) == '$' &&
				(*(p+2) == 'B' || *(p+2) == '@')) {
			kanji = 1;
			p += 3;
			continue;
		}
		if (*p == 0x1b && *(p+1) == '(' &&
				(*(p+2) == 'B' || *(p+2) == 'J')) {
			kanji = 0;
			p += 3;
			continue;
		}

		if (kanji) {
			code = *p * 256 + *(p + 1);
			p += 2;
		} else {
#ifndef USE_XDRAWSTRING
#ifdef USE_XDRAWSTRING_ONLY_SMALL
			if (char_size > 25) {
#endif /* USE_XDRAWSTRING_ONLY_SMALL */

			if (*p == ' ') {
				string_width += char_size / 2;
				p ++;
				continue;
			}
			code = ctable[*p++];

#ifdef USE_XDRAWSTRING_ONLY_SMALL
			} else {
				char_len = draw_onechar_x(p, string_width, 0);
				string_width += HORIZ_STEP(char_size, char_len);
				p ++;
				continue;
			}
#endif /*USE_XDRAWSTRING_ONLY_SMALL */
#else
			char_len = draw_onechar_x(p, string_width, 0);
			string_width += HORIZ_STEP(char_size, char_len);
			p ++;
			continue;
#endif /* end of USE_XDRAWSTRING */
		}

		char_len = draw_onechar(code, string_width, 0, char_size,
			char_size, page);
		string_width += HORIZ_STEP(char_size, char_len);
	}
}

static u_int
draw_onechar(code, x, y, width, height, page)
	u_int code;
	u_int x, y, width, height, page;
{
	struct vfont *vfc;
	XImage *xim = NULL;

	vfc = vfc_get(code, width, height, 1);

	if (no_bg_image) {
#ifdef ANTIALIAS
		if (!(mgp_flag & FL_NOALIAS))
			xim = vfc_image(vfc, fore_color, back_color);
#endif
		if (xim == NULL)
			xim = vfc->image;
		XPutImage(display, pixmap, gcfb, xim,
			vfc->minx, 0, x, y, vfc->charlen, vfc->height);
		if (xim != vfc->image)
			XDestroyImage(xim);
	} else {
		XPutImage(display, pixmap, gcfore, vfc->image,
			vfc->minx, 0, x, y, vfc->charlen, vfc->height);
		if (fore_color != depth_mask)
			XPutImage(display, maskpix, gcall, vfc->image,
				vfc->minx, 0, x, y, vfc->charlen, vfc->height);
	}
	return vfc->charlen;
}

#else /* VFLIB */

static void
draw_xstring(data, init)
	u_char *data;
	int init;
{
	u_char *p;
	u_int char_len;

	p = data;
	if (init) string_width = 0;

	while (*p != '\n' && *p != 0) {
		char_len = draw_onechar_x(p, string_width, 0);
		string_width += HORIZ_STEP(char_size, char_len);
		p ++;
	}
}
#endif /* end of VFLIB */

static void
draw_bar(cp)
	struct ctrl *cp;
{
	u_int width, swidth, st, len;
	Colormap cmap;
	Visual *visual;
	XColor col, scol;
	static GC gcbar, gcsbar;
	static u_long prevcolor = -1;

	if (!gcbar) {
		gcbar = XCreateGC(display, window, 0, 0);
		XSetFunction(display, gcbar, GXcopy);
		gcsbar = XCreateGC(display, window, 0, 0);
		XSetFunction(display, gcsbar, GXcopy);
	}
	col.pixel = cp->ctb_color;
	if (col.pixel == -1)
		col.pixel = fore_color;
	if (col.pixel != prevcolor) {
		prevcolor = col.pixel;
		visual = DefaultVisual(display, screen);
		cmap = DefaultColormap(display, screen);
		col.flags = DoRed|DoGreen|DoBlue;
		XQueryColor(display, cmap, &col);
		scol.red   = col.red   / 2;
		scol.green = col.green / 2;
		scol.blue  = col.blue  / 2;
		if (!XAllocColor(display, cmap, &scol))
			scol.pixel = col.pixel;
		XSetForeground(display, gcbar, col.pixel);
		XSetForeground(display, gcsbar, scol.pixel);
	}
	if (again) {
		pos_x = again_pos_x;
		pos_y = last_pos_y;
		again = 0;
	}
	width = cp->ctb_width * window_height / 1000;
	swidth = width / 2;
	width -= swidth;
	st = cp->ctb_start * window_width / 100;
	len = cp->ctb_length * window_width / 100;
	XFillRectangle(display, window, gcbar, st, pos_y, len, width);
	XFillRectangle(display, window, gcsbar, st, pos_y + width, len, swidth);

	last_pos_y = pos_y;
	last_pos_x = st + len;
	again_pos_x = pos_x;
	pos_y += width + swidth + VERT_GAP(char_size) / 2;
}

static void
process_direc(cp)
	struct ctrl *cp;
{
	switch(cp->ct_op) {
	case CTL_SIZE:
		char_size = window_height * cp->cti_value / 100;
		break;

	case CTL_VGAP:
		vert_gap = cp->cti_value;
		break;

	case CTL_HGAP:
		horiz_gap = cp->cti_value;
		break;

	case CTL_QUALITY:
	        if(!quality_flag) b_quality = cp->cti_value;
		break;

	case CTL_PAUSE:
		if (pause_line != line) {
			pause_flag = 1;
			pause_line = line;
		}
		break;

	case CTL_AGAIN:
		again = 1;
		break;

	case CTL_FORE:
		fore_color = cp->ctl_value;
		XSetForeground(display, gcfore, fore_color);
		XSetForeground(display, gcfb, fore_color);
		break;

	case CTL_BACK:
		back_color = cp->ctl_value;
		XSetForeground(display, gcback, back_color);
		XSetBackground(display, gcfb, back_color);
		bg_image = NULL;
		break;

	case CTL_CCOLOR:
		ctrl_color = cp->ctl_value;
		break;

	case CTL_CENTER:
		align = AL_CENTER;
		break;

	case CTL_LEFT:
		align = AL_LEFT;
		break;

	case CTL_RIGHT:
		align = AL_RIGHT;
		break;

	case CTL_CONT:
		cont = 1;
		break;

#ifdef VFLIB
	case CTL_VFONT:
		vfc_setfont(cp->ctc_value);
		break;
#endif /*VFLIB*/

	case CTL_XFONT:
		strcpy(xfont, cp->ctc_value);
		break;

	case CTL_BAR:
		draw_bar(cp);
		break;

	case CTL_IMAGE:
		image_load(cp->ctm_fname, cp->ctm_numcolor,
			cp->ctm_ximagesize, cp->ctm_yimagesize, 0,
			cp->ctm_zoomflag);
		break;

	case CTL_BIMAGE:
		if (mgp_flag & FL_BIMAGE)
			break;
		bg_image = cp;	/*update later*/
		break;

	case CTL_BGRAD:
		if (mgp_flag & FL_BIMAGE)
			break;
		bg_image = cp;	/*update later*/
		break;

	case CTL_LCUTIN:
		special = SP_LCUTIN;
		break;

	case CTL_RCUTIN:
		special = SP_RCUTIN;
		break;

	case CTL_SHRINK:
		special = SP_SHRINK;
		break;

	case CTL_PREFIX:
		prefix = cp->ctc_value;
		break;

	default:
		break;
	}
}

static void
back_gradation(cg0)
    struct ctrl_grad *cg0;
{
    Image *myimage, *image;
    Pixmap mypixmap;
    XImageInfo *ximageinfo;
    Visual *visual;
    byte * pic;
    int verbose = mgp_flag & FL_VERBOSE;
    static Cursor curs;
    u_int depth = DEFAULT_GRADDEPTH;
    u_int width, height;
    int dir, numcolor, zoomflag;
    int xzoomrate, yzoomrate;
    struct ctrl_grad cg1;
    struct ctrl_grad *cg;

    cg = &cg1;
    cg1 = *cg0;

    height= cg->ct_height;
    width = cg->ct_width;
    dir   = cg->ct_direction;
    numcolor = cg->ct_numcolor;
    zoomflag = cg->ct_zoomflag;

    screen = DefaultScreen(display);

    if (!curs)
	curs = XCreateFontCursor(display, XC_watch);
    XDefineCursor(display, window, curs); XFlush(display);

    /* make gradation image */
    pic = draw_gradation(cg);
    myimage = make_XImage(pic, width, height);

    if (numcolor < 64)
	myimage = reduce(myimage, numcolor, verbose);


    xzoomrate = cg->ct_xzoom;
    yzoomrate = cg->ct_yzoom;

    /* ??? */
    if (xzoomrate == 0) xzoomrate =100;
    if (yzoomrate == 0) yzoomrate =100;
    /* ??? */

#ifdef DEBUG
    fprintf(stderr, "Draw:zoomrate:(%d,%d)->(%d,%d)\n",
	    cg->ct_xzoom,cg->ct_yzoom,xzoomrate,yzoomrate);
    fprintf(stderr, "zoom mode %d: (%d,%d)->(%d,%d)[%d]\n",zoomflag,width,height,window_width,window_height,b_quality);
#endif

    if ( xzoomrate == 100 && yzoomrate == 100 ){
	if (verbose) fprintf(stderr, "no need zooming,skip\n");
    }else{
	image = myimage;
	myimage = zoom(image, xzoomrate, yzoomrate, verbose);
	freeImage(image);
#ifdef DEBUG
	fprintf(stderr, "zooming (%d,%d)->(%d,%d):*(%d,%d)\n",
		width, height, window_width, window_height,
		xzoomrate, yzoomrate);
#endif
    }


    /* check visual class */
    bestVisual(display, screen, myimage, &visual, &depth);
    if (verbose && (visual != DefaultVisual(display, screen)))
	printf("  Using %s visual\n", nameOfVisualClass(visual->class));

    if (! (ximageinfo= imageToXImage(display, screen, visual,
				     depth, myimage, 0, 0, verbose))) {
	fprintf(stderr, "Cannot convert Image to XImage\n");
	exit(1);
    }

    if (((mypixmap= 
	  ximageToPixmap(display, DefaultRootWindow(display), ximageinfo))
	 == None) && verbose) {
	fprintf(stderr, "Cannot create image in server\n");
	exit(1);
    }

    /* draw background */
    XSetWindowBackgroundPixmap(display, window, mypixmap);

    /* finish */
    XFreePixmap(display, mypixmap);
    freeXImage(myimage, ximageinfo);
    freeImage(myimage);
    XUndefineCursor(display, window); XFlush(display);
}

static void
image_load(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
	char *filename;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;
{
	Image *image, *myimage;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	Visual *visual;
	u_int depth;
	int width, height;
	int xzoomrate, yzoomrate;
	int	verbose = mgp_flag & FL_VERBOSE;
	static Cursor curs;

	if (pause_flag)
		return;

	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, window, curs); XFlush(display);

	screen = DefaultScreen(display);

	if ((myimage = loadImage(filename, verbose)) == NULL)
		exit(-1);	/* fail to load image data */
	width = myimage->width;
	height = myimage->height;

	if (myimage->depth == 1) {
		XColor xc;
	        Colormap cmap;

	        cmap = DefaultColormap(display, screen);
		xc.flags = DoRed | DoGreen | DoBlue;
		xc.pixel = fore_color;
		XQueryColor(display, cmap, &xc);
		*(myimage->rgb.red + 1) = xc.red;
		*(myimage->rgb.green + 1) = xc.green;
		*(myimage->rgb.blue + 1) = xc.blue;
		xc.pixel = back_color;
		XQueryColor(display, cmap, &xc);
		*myimage->rgb.red = xc.red;
		*myimage->rgb.green = xc.green;
		*myimage->rgb.blue = xc.blue;
	}

	if (numcolor)
		myimage = reduce(myimage, numcolor, verbose);

	if (zoomflag == 2) {
		if (ximagesize) {
			if (ximagesize == window_width)
				ximagesize = 100;
			else
				ximagesize = window_width * 100 / ximagesize;
		}
		if (yimagesize) {
			if (yimagesize == window_height)
				yimagesize = 100;
			else
				yimagesize = window_height * 100 / yimagesize;
		}
	}
	if (ximagesize != 0 && yimagesize != 0) {
		image = myimage;
		if (!zoomflag) {
			xzoomrate = window_width * ximagesize / width;
			yzoomrate = window_height * yimagesize / height;
		} else {
			xzoomrate = ximagesize;
			yzoomrate = yimagesize;
		}
		if (xzoomrate != 100 || yzoomrate != 100) {
			myimage = zoom(image, xzoomrate, yzoomrate, verbose);
			freeImage(image);
		}
	} else
		xzoomrate = yzoomrate = 100;

	bestVisual(display, screen, myimage, &visual, &depth);
	if (verbose && (visual != DefaultVisual(display, screen)))
		printf("  Using %s visual\n", nameOfVisualClass(visual->class));

	if (! (ximageinfo= imageToXImage(display, screen, visual, depth, myimage,
		0, 0, verbose))) {
			fprintf(stderr, "Cannot convert Image to XImage\n");
			exit(1);
	}

	if (((mypixmap= ximageToPixmap(display,
		DefaultRootWindow(display), ximageinfo)) == None) && verbose) {
			fprintf(stderr, "Cannot create image in server\n");
			exit(1);
	}

	if (backflag) {
		XSetWindowBackgroundPixmap(display, window, mypixmap);
		goto end;
	}

	switch (align) {
	case AL_LEFT:
		pos_x = 0;
		break;
	case AL_CENTER:
		pos_x = (window_width - width * xzoomrate /100) / 2;
		break;
	case AL_RIGHT:
		pos_x = window_width - width * xzoomrate /100;
		break;
	}

	if (cont) {
		pos_x = last_pos_x;
		pos_y = last_pos_y;
		cont = 0;
	}
	if (again) {
		pos_x = again_pos_x;
		pos_y = last_pos_y;
		again = 0;
	}

	XCopyArea(display, mypixmap, window, gcall, 0, 0,
		width * xzoomrate / 100,
		height * yzoomrate / 100, pos_x, pos_y);

	last_pos_y = pos_y;
	last_pos_x = pos_x + width * xzoomrate / 100;
	again_pos_x = pos_x;
	pos_y += height * yzoomrate / 100;

end:
	XFreePixmap(display, mypixmap);
	freeXImage(myimage, ximageinfo);
	freeImage(myimage);
	XUndefineCursor(display, window); XFlush(display);
}

static u_int
draw_onechar_x(p, x, y)
	u_char *p;
	u_int x, y;
{
	u_char fontstring[BUFSIZ];
	static u_int last_char_size = 0;
	static u_char last_xfont[BUFSIZ] = "";
	u_int i;
	u_int char_len;

	if (strcmp(xfont, last_xfont) || char_size != last_char_size) {
		/* load new font */
		if (xfontstruct)
			XFreeFont(display, xfontstruct);

#if 1
		/*
		 * Look for the best font possible.
		 * 1. Check for a font that is smaller than the required one.
		 *    By using smaller font, we won't make the screen garbled.
		 * 2. If 1. is impossible, look for slightly larger font than
		 *    the required one.
		 */
#define CHKFONTSIZE(x)	((char_size < x) ? x : (char_size - x))
#define	FONTSEARCH	20
#else
		/* The original one. */
#define CHKFONTSIZE(x)	((int)char_size - x)
#define	FONTSEARCH	20
#endif
		for (i = 0; i < FONTSEARCH; i++) {
			if (CHKFONTSIZE(i) < 0)
				continue;
			sprintf(fontstring, FONT_FORMAT, xfont, CHKFONTSIZE(i));
			if (xfontstruct = XLoadQueryFont(display, fontstring))
				break;
		}
#undef CHKFONTSIZE
#undef FONTSEARCH
		if (xfontstruct == NULL) {
			sprintf(fontstring, FONT_FORMAT, xfont, char_size);
			fprintf(stderr, "can't load font %s "
				"(nor font in similar size)\n", fontstring);
			exit(-1);
		}
		XSetFont(display, gcfore, xfontstruct->fid);
		XSetFont(display, gcfb, xfontstruct->fid);
		XSetFont(display, gcall, xfontstruct->fid);

		last_char_size = char_size;
		strcpy(last_xfont, xfont);
	}

	char_len = XTextWidth(xfontstruct, p, 1);
	if (no_bg_image) {
		XDrawString(display, pixmap, gcfb,
			x, y + char_size - xfontstruct->descent / 2 - 1, p, 1);
	} else {
		XDrawString(display, pixmap, gcfore,
			x, y + char_size - xfontstruct->descent / 2 - 1, p, 1);
		if (fore_color != depth_mask)
			XDrawString(display, maskpix, gcall,
				x, y + char_size - xfontstruct->descent / 2 - 1, p, 1);
	}

	return char_len;
}

static void
lcutin(data, page)
	u_char *data;
	u_int page;
{
	u_int	lx;
	u_int	step, x;

#ifdef VFLIB
	draw_vfstring(data, 1, page);
#else
	draw_xstring(data, 1);
#endif

	lx = set_position();
	step = lx / 20;

	for (x = 0; x < lx + step; x += step) {
		if (x > lx) x = lx;
		if (no_bg_image) {
			XCopyArea(display, pixmap, window, gcall,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, char_size));
		} else {
			if (fore_color != depth_mask)
				XCopyArea(display, maskpix, window, gcmask,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, char_size));
			XCopyArea(display, pixmap, window, gcor,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, char_size));
		}
		XFlush(display);
		usleep(CUTIN_DELAY);
		XClearArea(display, window, x, VERT_POS(pos_y, char_size),
				string_width, char_size, 0);
	}
}

static void
rcutin(data, page)
	u_char *data;
	u_int page;
{
	int	lx, fx;
	int	step, x;

#ifdef VFLIB
	draw_vfstring(data, 1, page);
#else
	draw_xstring(data, 1);
#endif
	lx = set_position();
	fx = window_width - string_width;
	step = (fx - lx) / 20;

	for (x = fx; x > lx - step; x -= step) {
		if (x < lx) x = lx;
		if (no_bg_image) {
			XCopyArea(display, pixmap, window, gcall,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, char_size));
		} else {
			if (fore_color != depth_mask)
				XCopyArea(display, maskpix, window, gcmask,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, char_size));
			XCopyArea(display, pixmap, window, gcor,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, char_size));
		}
		XFlush(display);
		usleep(CUTIN_DELAY);
		XClearArea(display, window, x, VERT_POS(pos_y, char_size),
				string_width, char_size, 0);
	}
}

static void
shrink(data, page)
	u_char *data;
	u_int page;
{
	int min_csize = char_size;
	int max_csize = window_height / 4;
	int csize, i, x;
	u_int step = (max_csize - min_csize) / 3;

	if (!step)
		step = 1;

	if (align != AL_CENTER) {
		fprintf(stderr, "align is not center: \n");
		return;
	}

		if (char_size < min_csize)
			char_size = min_csize;

	csize = char_size;
	for (i = max_csize; i > min_csize; i -= step) {
		char_size = i;
#ifdef VFLIB
		draw_vfstring(data, 1, page);
#else
		draw_xstring(data, 1);
#endif
		x = (window_width - string_width) / 2;
		if (no_bg_image) {
			XCopyArea(display, pixmap, window, gcall,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, csize));
		} else {
			if (fore_color != depth_mask)
				XCopyArea(display, maskpix, window, gcmask,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, csize));
			XCopyArea(display, pixmap, window, gcor,
					0, 0, string_width, char_size,
					x, VERT_POS(pos_y, csize));
		}
		XFlush(display);
		usleep(SHRINK_DELAY);
		if (no_bg_image) {
			XFillRectangle(display, pixmap, gcback,
				0, 0, window_width, char_size);
		} else {
			XFillRectangle(display, pixmap, gcnone,
				0, 0, window_width, char_size);
			if (fore_color != depth_mask)
				XFillRectangle(display, maskpix, gcnone,
					0, 0, window_width, char_size);
		}
		XClearArea(display, window, x, VERT_POS(pos_y, csize),
			string_width, char_size, 0);
	}
	char_size = csize;
}

static int
set_position()
{
	u_int	x;

	switch(align) {
	case AL_CENTER:
		x = (window_width - string_width)/ 2;
		break;

	case AL_LEFT:
		x = 0;
		break;

	case AL_RIGHT:
		x = window_width - string_width;
		break;
	}

	return x;
}

static void timebar(page)
	u_int page;
{
	int pos, n, p, barlen;
	GC *pgc;

	if (t_start == 0 || tbar_flag == 0)
		return;

	pos =  (window_width - 2) * (page - 1) / (maxpage - 1);
	p = (time(0) - t_start) * 100;
	barlen = window_width - window_width * p / t_fin / 6000;

	if (window_width * 50 / 100 < barlen)
		pgc = &gcgreen;
	else if (window_width * 30 / 100 < barlen)
		pgc = &gcyellow;
	else
		pgc = &gcred;
			
	if (barlen > 0) {
		XFillRectangle(display, window, gcback,
			0, window_height - 2, window_width, window_height);
		XFillRectangle(display, window, *pgc,
			window_width - barlen, window_height - 1,
			window_width, window_height);
		XFillRectangle(display, window, *pgc,
			pos, window_height - 5,
			2, 5);
	} else if (barlen < 0) {
		barlen = - barlen;
		n = p / t_fin / 6000;
		if (n > window_height - 1)
			n = window_height - 1;
		if (n)
			XFillRectangle(display, window, gcred,
				0, window_height - n,
				barlen, window_height);
		XFillRectangle(display, window, gcback,
			0, window_height - n - 2,
			window_width, window_height - n);
		XFillRectangle(display, window, gcred,
			0, window_height - n - 1,
			barlen % window_width, window_height - n);
		XFillRectangle(display, window, gcred,
			pos, window_height - n - 1 - 5,
			2, 5);
	}
}
