/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: parse.c,v 1.52 1997/12/24 06:41:42 yuo Exp $
 */

#include "mgp.h"

/*#define PARSE_DEBUG*/

static struct ctrl *parse_directive __P((u_char *, int *, u_int));
static void read_file __P((FILE *));
static struct ctrl *set_control __P((u_int, u_char, u_char **, u_int));
static u_char search_ctltype __P((u_char *));
#ifdef PARSE_DEBUG
void debug __P((void));
#endif
static int default_check __P((u_char **, u_char *));
static u_char *allocpy __P((u_char *));
static int chkfile __P((u_char *));
static struct ctrl *ctllastitem __P((struct ctrl *));
static void ctlappend __P((struct ctrl *, struct ctrl *));
static void ctlinsert __P((struct ctrl **, struct ctrl *));
static struct ctrl *ctlalloc1 __P((u_char));
static void ctlfree __P((struct ctrl *));
static struct ctrl *ctlcopy __P((struct ctrl *));
static FILE *fsearchopen __P((char *, char *, char **));

/*image*/
extern int findImage __P((char *, char *));

#ifndef HAVE_STRSEP
char *
strsep(stringp, delim)
	char **stringp;
	register const char *delim;
{
	register char *s = *stringp;
	register char *dl;
	char c, d;
	char *ret;

	if (s == NULL)
		return s;
	for (ret = s; c = *s; s++) {
		for (dl = delim; d = *dl; dl++) {
			if (c == d) {
				*s = '\0';
				*stringp = s + 1;
				return ret;
			}
		}
	}
	*stringp = NULL;
	return ret;
}
#endif

void
load_file(filename)
	u_char *filename;
{
	FILE *fp;

	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "can't open %s\n", filename);
		exit(-1);
	}
	read_file(fp);
	fclose(fp);
#ifdef PARSE_DEBUG
	debug();
	exit(0);
#endif
}

void
load_include_file(filename)
	u_char *filename;
{
	FILE *fp;
	static char *searchpath[] = {
		"",	/*mgp_fname*/
#ifdef MGPLIBDIR
		MGPLIBDIR "/",
#endif
		NULL,
	};

	searchpath[0] = mgp_fname;
	fp = fsearchopen(filename, "r", searchpath);
	if (!fp) {
		fprintf(stderr, "can't open %s\n", filename);
		exit(-1);
	}
	read_file(fp);
	fclose(fp);
#ifdef PARSE_DEBUG
	debug();
	exit(0);
#endif
}

static struct ctrl *
parse_directive(p, pagebreak, page)
	u_char *p;
	int *pagebreak;
	u_int page;
{
	u_char **direcp, *direc[MAXDIREC];
	u_char **argp, *arg[MAXARG];
	u_int direc_cnt, arg_cnt;
	struct ctrl sentinel;
	struct ctrl *cp;
	u_char op;
	int i;

	assert(p);

	arg_cnt = direc_cnt = 0;
	cp = &sentinel;
	if (pagebreak)
		*pagebreak = 0;

	/*
	 * Divide the line into each directive
	 */
	for (direcp = direc;
	     p && (*direcp = strsep((char **)&p, ",")) != NULL;
	     /*nothing*/) {
		if (**direcp != '\0') {
			direcp++;
			direc_cnt++;
		}
	}

	/*
	 * Analyze arguments in each directive
	 */
	for (i = 0; i < direc_cnt; i++) {
		p = direc[i];
		arg_cnt = 0;
		for (argp = arg;
		     p && (*argp = strsep((char **)&p, " ")) != NULL;
		     /*nothing*/) {
			if (**argp != '\0') {
				argp++;
				arg_cnt ++;
			}
		}
		op = search_ctltype(arg[0]);
		if (op == CTL_PAGE) {
			if (pagebreak)
				*pagebreak = 1;
		}
		cp->ct_next = set_control(page, op, arg, arg_cnt);
		if (cp->ct_next)
			cp = cp->ct_next;
	}

	return sentinel.ct_next;
}

static void
read_file(fp)
	FILE *fp;
{
	u_char buf[BUFSIZ], *p;
	u_int line;
	u_int page;
	u_int page_line_count = 0;
	int def_num;
	u_char op;
	struct ctrl **ch;
	struct ctrl *cp;

	page = 1;
	line = 0;

	/*
	 * default analysis in preamble
	 */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (buf[strlen(buf) - 1])
			buf[strlen(buf) - 1] = '\0';

		if (buf[0] == '#')
			continue;
		if (buf[0] != '%') {
			fprintf(stderr, "no text allowed in preamble\n");
			continue;
		}

		p = buf + 1;
		def_num = default_check(&p, &op);
		if (op == CTL_PAGE)
			goto page;

		if (op == CTL_INCLUDE)
			load_include_file(p);

		if (def_num == -1)
			continue;

		switch (op) {
		case CTL_DEFAULT:
			ch = &default_control[def_num];
			break;
		case CTL_TAB:
			ch = &tab_control[def_num];
			break;
		default:
			continue;
		}

		cp = parse_directive(p, NULL, page);

		if (cp) {
			if (!*ch)
				*ch = cp;
			else
				ctlappend(*ch, cp);
		}
	}

	/*
	 * page analysis
	 */
page:
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (buf[strlen(buf) - 1])
			buf[strlen(buf) - 1] = '\0';

		if (buf[0] == '#')
			continue;
		else if (buf[0] == '%') {
			/*
			 * this is directive
			 */
			int pb;

			pb = 0;
			cp = parse_directive(buf + 1, &pb, page);
			if (pb) {
				/* seen pagebreak */
				page_attribute[page].pg_linenum = line;
				page++;
				line = 0;
				continue;
			}

			if (cp) {
				/*
				 * append to the currently existing
				 * page struct.
				 */
				ch = &page_control[page][line];
				if (!*ch)
					*ch = cp;
				else
					ctlappend(*ch, cp);
			}
		} else {
			/* this is data */
			page_data[page][line] = allocpy(buf);
			line++;
			page_line_count = 0;
		}
	}

	/* treat as we've seen pagebreak */
	page_attribute[page].pg_linenum = line;
#if 0
	page++;
	line = 0;
#endif

	maxpage = page;
}

static struct ctrl *
set_control(page, op, arg, arg_cnt)
	u_int page;
	u_char op;
	u_char **arg;
	u_int arg_cnt;
{
	struct ctrl *cp;
	u_char *p;
	u_char buf[BUFSIZ];
	int i, colors;
	int gradsize;
	float zoomrate;
	int width, height;
	int base_width, base_height;
	int draw_width, draw_height;

	cp = ctlalloc1(op);
	cp->ct_page = page;

	switch(op) {
	case CTL_LEFT:
	case CTL_RIGHT:
	case CTL_CENTER:
	case CTL_SHRINK:
	case CTL_LCUTIN:
	case CTL_RCUTIN:
	case CTL_CONT:
	case CTL_NODEF:
	case CTL_PAUSE:
	case CTL_AGAIN:
		/* no arg */
		if (op == CTL_NODEF)
			page_attribute[page].pg_flag |= PGFLAG_NODEF;
		break;

	case CTL_SIZE:
	case CTL_HGAP:
	case CTL_VGAP:
	case CTL_QUALITY:
		/* one int arg */
		cp->cti_value = atoi(arg[1]);
		break;

	case CTL_FORE:
	case CTL_BACK:
	case CTL_CCOLOR:
		/* one long arg */
		if (get_color(arg[1], &cp->ctl_value) < 0) {
			fprintf(stderr, "Can't allocate color %s\n", arg[1]);
			parse_error++;
		}
		break;

	case CTL_BGRAD:
	     /*  width:  gradation cell size (x-%)
	      *  height: gradation cell size (y-%)
	      *  dir:    gradation direction (degree)
	      *  color:  color list (max 16 colors)
	      */

	     /* set default values */
	     cp->ctd_width  = window_width;
	     cp->ctd_height = window_height;
	     cp->ctd_dir = 0;
	     cp->ctd_numcolor = DEFAULT_GRADCOLORS;
	     cp->ctd_zoomflag = 1;
	     cp->ctd_depth = 8;

	     /*
	      * here we write down each cases, just for readability.
	      */

	     switch (arg_cnt) {
	     case 1:
		 width=height=100;
		 break;
	     case 2:
		 width = atoi(arg[1]);
		 break;
	     case 3:
		 width  = atoi(arg[1]);
		 height = atoi(arg[2]);
		 break;
	     case 4:
		 width   = atoi(arg[1]);
		 height  = atoi(arg[2]);
		 cp->ctd_numcolor= atoi(arg[3]);
		 break;
	     case 5:
		 width   = atoi(arg[1]);
		 height  = atoi(arg[2]);
		 cp->ctd_numcolor= atoi(arg[3]);
		 if(*arg[4] != '-'){
		     cp->ctd_dir     = atoi(arg[4]);
		 }else{
		     ++arg[4];
		     cp->ctd_dir     = -1 * atoi(arg[4]);
		 }
		 break;
	     case 6:
	     default:
		 width   = atoi(arg[1]);
		 height  = atoi(arg[2]);
		 cp->ctd_numcolor= atoi(arg[3]);
		 if(*arg[4] != '-'){
		     cp->ctd_dir     = atoi(arg[4]);
		 }else{
		     ++arg[4];
		     cp->ctd_dir     = -1 * atoi(arg[4]);
		 }
		 cp->ctd_zoomflag= atoi(arg[5]);
		 break;
	     }

	     /* set color list */
	     cp->ct_val.ctrl_grad.colors[0] = 
		 (struct g_color *)name2gcolor(DEFAULT_GRADSTART);
	     cp->ct_val.ctrl_grad.colors[1] = 
		 (struct g_color *)name2gcolor(DEFAULT_GRADEND);
	     cp->ctd_g_colors = 2;

	     if (arg_cnt < 7) {
		 /* use default gradation color set*/
	     } else if (arg_cnt == 7) {
		 cp->ct_val.ctrl_grad.colors[0] = 
		     (struct g_color *)name2gcolor(arg[i]);
	     } else if (arg_cnt > 7) {
		 colors = 0;
		 for (i = 6; i < arg_cnt; i++){
		     cp->ct_val.ctrl_grad.colors[colors] = 
			 (struct g_color *)name2gcolor(arg[i]);
		     colors++;
		 }
		 cp->ctd_g_colors = colors;
	     }

	     /* check args rang and canonicarize */

	     /* normalize direction */
	     /* if direction is negative, set to nonliner mode */
	     if (cp->ctd_dir < 0){
		 /* set to circle gradation */
		 cp->ctd_mode = 1;
		 cp->ctd_dir = abs(cp->ctd_dir);
	     } else {
		 /* set to liner gradation */
		 cp->ctd_mode = 0;
	     }

	     while (cp->ctd_dir >= 360) cp->ctd_dir -= 360;

	     /* ------------ quality ajasting -------------------- */
	     /* base window size */
	     if (b_quality == 100){
		 base_width = window_width;
		 base_height= window_height;
	     } else {
		 base_width  = (int)(b_quality * window_width / 100.0);
		 base_height = (int)(b_quality * window_height / 100.0);
	     }

	     /* draw window size */

	     /* 0 means a full window size */
	     if(width == 0) width = 100;
	     if(height == 0) height = 100;
		 
	     draw_width = (int)(base_width * width / 100.0);
	     draw_height= (int)(base_height * height / 100.0);
	     
	     /* set size to drawing stracture */
	     cp->ctd_width = draw_width;
	     cp->ctd_height = draw_height;

#ifdef GRAD_DEBUG
	     fprintf(stderr, "parse: Base(%d,%d):(%d,%d)->(%d,%d)\n",
		     base_width,base_height,width,height,
		     cp->ctd_width,cp->ctd_height);
#endif

 	     /* performace enhance hack for special case  */
	     switch (cp->ctd_dir){
	     case 0:	
	     case 180:	
		 /* force set to generate 4dot-width image */
		 cp->ctd_width  = 4;
		 break;
	     case 90:	
	     case 270:	
		 /* force set to generate 4dot-height image */
		 cp->ctd_height  = 4;
		 break;
	     default:
		 break;
	     }
	     
	     /* if zoomflag is on, zoom to fullsize.
	      *  else, zoom to basewindow size.
	      */
	     if(cp->ctd_zoomflag == 1){
		 cp->ctd_xzoom = 
		   (int)((window_width *100.0) / (float)cp->ctd_width);
		 cp->ctd_yzoom =
		   (int)((window_height*100.0) / (float)cp->ctd_height);
	     }else{
		 cp->ctd_xzoom = (int)(1.0/(float)b_quality);
		 cp->ctd_yzoom = (int)(1.0/(float)b_quality);
	     }
#ifdef GRAD_DEBUG
	     fprintf(stderr,"Parse:zoom mode:%d,[%d,%d]\n",
		     cp->ctd_zoomflag,cp->ctd_xzoom,cp->ctd_yzoom);
#endif
	     
#if 0
#define MAX_GRAD_IMAGE_SIZE  120000
	     /* rotation calcilation is very slow,
	      * therefore, minimaize image size 
	      */
	     if ( cp->ctd_width == window_width &&
		  cp->ctd_height== window_height){
		 /* default: generae 1/4*1/4 size image */
		 cp->ctd_width = (int)(window_width/4);
		 cp->ctd_height = (int)(window_height/4);
		 cp->ctd_zoomflag = 1;
	     } else {
		 gradsize = cp->ctd_width * cp->ctd_height;
		 if ( gradsize > MAX_GRAD_IMAGE_SIZE){
		     if (cp->ctd_zoomflag){
			 zoomrate = sqrt(MAX_GRAD_IMAGE_SIZE / gradsize);
			 cp->ctd_width = (int)(cp->ctd_width * zoomrate);
			 cp->ctd_height= (int)(cp->ctd_height * zoomrate);
		     }
		 }
	     }
#endif
	     /* other optimization / canonicalization TBD */

	     break;

	case CTL_XFONT:
		if (strncmp(arg[1], "medium", 6) == 0 ||
		    strncmp(arg[1], "bold", 4) == 0) {
			/* for backward compatibility */
			sprintf(buf, "times-%s", arg[1]);
			cp->ctc_value = allocpy(buf);
		} else
			cp->ctc_value = allocpy(arg[1]);
		break;

	case CTL_BIMAGE:
	case CTL_VFONT:
	case CTL_PREFIX:
		/* one char arg */
		cp->ctc_value = allocpy(arg[1]);

		if (op == CTL_BIMAGE) {
			strcpy((char *)page_attribute[page].pg_bimage,
			       cp->ctc_value);
			chkfile(cp->ctc_value);
		}

		if (op == CTL_PREFIX) {
			char *p = cp->ctc_value;

			/* prefix replacements */
			while (*p) {
				switch (*p) {
				case '_':
					*p = ' ';
					break;
				case '.':
					*p = '\0';
					break;
				}
				p++;
			}
		}
		break;

	case CTL_IMAGE:
		/* image */
		cp->ctm_numcolor = 0;
		cp->ctm_ximagesize = 0;
		cp->ctm_yimagesize = 0;
		cp->ctm_zoomflag = 0;
		/* handle special case */
		for (i = 2; i < arg_cnt; i++) {
			if ((p = strchr(arg[i], 'x'))) {
				cp->ctm_zoomflag = 2;
				cp->ctm_ximagesize = atoi(arg[i]);
				cp->ctm_yimagesize = atoi(p + 1);
				arg_cnt = i > 3 ? 3 : i;
				break;
			}
		}
		switch (arg_cnt) {
		default:
		case 6:
			cp->ctm_zoomflag = atoi(arg[5]) ? 1 : 0;
			/*FALLTHRU*/
		case 5:
			cp->ctm_yimagesize = atoi(arg[4]);
			/*FALLTHRU*/
		case 4:
			cp->ctm_ximagesize = atoi(arg[3]);
			if (arg_cnt == 4)
				cp->ctm_yimagesize = cp->ctm_ximagesize;
			/*FALLTHRU*/
		case 3:
			cp->ctm_numcolor = atoi(arg[2]);
			/*FALLTHRU*/
		case 2:
			cp->ctm_fname = allocpy(arg[1]);
			break;
		case 1:
			fprintf(stderr, "file name is not specified in %%image\n");
			parse_error++;
			break;
		}
		chkfile(cp->ctm_fname);
		break;

	case CTL_BAR:
		/* bar [color [width%0 [start% [length%]]]] */
		cp->ctb_color = -1;
		cp->ctb_width = 10;
		cp->ctb_start = 0;
		cp->ctb_length = 100;
		switch (arg_cnt) {
		default:
		case 5:
			cp->ctb_length = atoi(arg[4]);
			if (cp->ctb_length < 0) cp->ctb_length = 0;
			if (cp->ctb_length > 100) cp->ctb_length = 100;
			/*FALLTHRU*/
		case 4:
			cp->ctb_start = atoi(arg[3]);
			if (cp->ctb_start < 0) cp->ctb_start = 0;
			if (cp->ctb_start > 100) cp->ctb_start = 100;
			if (cp->ctb_start + cp->ctb_length > 100)
				cp->ctb_length = 100 - cp->ctb_start;
			/*FALLTHRU*/
		case 3:
			cp->ctb_width = atoi(arg[2]);
			if (cp->ctb_width < 0) cp->ctb_width = 0;
			if (cp->ctb_width > 1000) cp->ctb_width = 1000;
			/*FALLTHRU*/
		case 2:
			if (get_color(arg[1], &cp->ctb_color) < 0) {
				fprintf(stderr, "Can't allocate color %s\n", arg[1]);
				parse_error++;
			}
			/*FALLTHRU*/
		case 1:
			break;
		}
		break;

	default:
		/* believe this is comment */
		/* do nothing */
		free(cp);
		cp = NULL;
		break;
	}

	return cp;
}

static u_char
search_ctltype(arg)
	u_char	*arg;
{
	u_int i = 0;

	while (ctl_words[i].ctl_strlen != 0) {
		if (!strncasecmp(arg, ctl_words[i].ctl_string,
				ctl_words[i].ctl_strlen)) {
			return ctl_words[i].ctl_type;
		}
		i ++;
	}
	return 0;
}

#ifdef PARSE_DEBUG
void
debug()
{
	int page, line, i;
	struct ctrl *cp;

	for (line = 0; line < MAXLINE; line ++) {
		i = 0;
		for (cp = default_control[line]; cp; cp = cp->ct_next) {
			printf(" def line %d i %d type %s(%d)\n", line, i,
				ctl_words[cp->ct_op].ctl_string, cp->ct_op);
			i++;
		}
	}
	for (page = 0; page < MAXPAGE; page ++) {
		for (line = 0; line < MAXLINE; line ++) {
			i = 0;
			for (cp = page_control[page][line]; cp;
					cp = cp->ct_next) {
				printf(" page %d line %d i %d type %s(%d)\n",
					page, line, i,
					ctl_words[cp->ct_op].ctl_string,
					cp->ct_op);
				i++;
			}
			if (page_data[page][line])
				printf("data [%s]\n", page_data[page][line]);
		}
	}
}
#endif /*PARSE_DEBUG*/

static int
default_check(buf, op)
	u_char **buf;
	u_char *op;
{
	u_char	*p = *buf;
	int i, number = 0;

	for (i = 0; ctl_words[i].ctl_strlen != 0; i++) {
		if (!strncasecmp(p, ctl_words[i].ctl_string,
				ctl_words[i].ctl_strlen)) {
			break;
		}
	}

	*op = ctl_words[i].ctl_type; /* may be 0 */
	p += ctl_words[i].ctl_strlen; /* may be 0 */

	while (isspace(*p))
		p++;

	while (isdigit(*p)) {
		number = number * 10 + *p - '0';
		p++;
	}

	while (isspace(*p))
		p++;
	*buf = p;

	return number - 1;
}

static u_char *
allocpy(p)
	u_char *p;
{
	u_char *q;

	q = (u_char *)malloc(strlen(p) + 1);
	return (u_char *)strcpy((char *)q, (char *)p);
}

static int
chkfile(p)
	u_char *p;
{
	char buf[BUFSIZ];

	buf[0] = '\0';
	if (findImage(p, buf) >= 0)
		return 0;
	fprintf(stderr, "File %s can't found\n", p);
	parse_error++;
	return -1;
}

/*------------------------------------------------------------*/

static struct ctrl *
ctllastitem(this)
	struct ctrl *this;
{
	struct ctrl *p;

	assert(this);
	p = this;
	while (p && p->ct_next)
		p = p->ct_next;

	return p;
}

static void
ctlappend(to, this)
	struct ctrl *to;
	struct ctrl *this;
{
	struct ctrl *p;

	assert(to);
	assert(this);
	p = ctllastitem(to);
	p->ct_next = this;
}

static void
ctlinsert(here, this)
	struct ctrl **here;
	struct ctrl *this;
{
	struct ctrl *p;
	struct ctrl *q;

	assert(here); assert(this);
	p = *here;
	*here = this;
	q = ctllastitem(this);
	q->ct_next = p;
}

static struct ctrl *
ctlalloc1(op)
	u_char op;
{
	struct ctrl *p;

	p = (struct ctrl *)malloc(sizeof(struct ctrl));
	if (!p) {
		perror("malloc");
		exit(-1);
	}
	memset(p, 0, sizeof(struct ctrl));
	p->ct_op = op;
	p->ct_next = NULL;	/*just to make sure*/
	return p;
}

static void
ctlfree(this)
	struct ctrl *this;
{
	struct ctrl *p;
	struct ctrl *q;

	assert(this);
	p = this;
	do {
		q = p->ct_next;
		free(p);
		p = q;
	} while (p);
}

static struct ctrl *
ctlcopy(this)
	struct ctrl *this;
{
	struct ctrl *dst0;
	struct ctrl *dst;
	struct ctrl *src;

	src = this;
	if (!src)
		return NULL;
	dst = dst0 = ctlalloc1(0);
	memcpy(dst, src, sizeof(struct ctrl));
	src = src->ct_next;
	while (src) {
		dst->ct_next = ctlalloc1(0);
		dst = dst->ct_next;
		memcpy(dst, src, sizeof(struct ctrl));
		src = src->ct_next;
	}

	return dst0;
}

int
ctlcmp(a, b)
	struct ctrl *a;
	struct ctrl *b;
{
	int i;

	assert(a);
	assert(b);

	if (a->ct_op != b->ct_op)
		return 1;
	if (a->ct_flag != b->ct_flag)
		return 1;
	switch (a->ct_op) {
	case CTL_NOOP:
	case CTL_LEFT:
	case CTL_CENTER:
	case CTL_RIGHT:
	case CTL_SHRINK:
	case CTL_LCUTIN:
	case CTL_RCUTIN:
	case CTL_CONT:
	case CTL_NODEF:
	case CTL_PAUSE:
	case CTL_AGAIN:
		return 0;
	case CTL_SIZE:
	case CTL_HGAP:
	case CTL_VGAP:
	case CTL_QUALITY:
		return (a->cti_value == b->cti_value) ? 0 : 1;
	case CTL_FORE:
	case CTL_BACK:
	case CTL_CCOLOR:
		return (a->ctl_value == b->ctl_value) ? 0 : 1;
	case CTL_XFONT:
	case CTL_VFONT:
	case CTL_BIMAGE:
	case CTL_PREFIX:
		return strcmp(a->ctc_value, b->ctc_value);
	case CTL_DEFAULT:
	case CTL_TAB:
	case CTL_PAGE:
	case CTL_INCLUDE:
		return 1;
	case CTL_IMAGE:
		if (a->ctm_numcolor == b->ctm_numcolor
		 && a->ctm_ximagesize == b->ctm_ximagesize
		 && a->ctm_yimagesize == b->ctm_yimagesize
		 && a->ctm_zoomflag == b->ctm_zoomflag)
			return strcmp(a->ctm_fname, b->ctm_fname);
		else
			return 1;
	case CTL_BGRAD:
		if (a->ctd_g_colors == b->ctd_g_colors
		 && a->ctd_numcolor == b->ctd_numcolor
		 && a->ctd_dir == b->ctd_dir
		 && a->ctd_width == b->ctd_width
		 && a->ctd_height == b->ctd_height
		 && a->ctd_zoomflag == b->ctd_zoomflag
		 && a->ctd_depth == b->ctd_depth)  {
			for (i = 0; i < a->ctd_g_colors; i++) {
				if (bcmp(a->ctd_colors[0], b->ctd_colors[0],
						sizeof(struct g_color)) != 0)
					return 1;
			}
			return 0;
		} else
			return 1;
	case CTL_BAR:
		if (a->ctb_color == b->ctb_color
		 && a->ctb_width == b->ctb_width
		 && a->ctb_start == b->ctb_start
		 && a->ctb_length == b->ctb_length)
			return 0;
		else
			return 1;
	default:
		assert(0);
	}
}

static FILE *
fsearchopen(fname, mode, path)
	char *fname;
	char *mode;
	char **path;
{
	FILE *fp;
	char buf[MAXPATHLEN];
	char *p;
	int i;

	i = -1;
	fp = NULL;
	while (1) {
		p = (i == -1) ? "" : path[i];
		if (!p)
			break;
		strcpy(buf, p);
		if (p = strrchr(buf, '/'))
			p[1] = '\0';
		strcat(buf, fname);
		fp = fopen(buf, mode);
		if (fp)
			break;
		i++;
	}
	return fp;
}
