/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: x11.c,v 1.12 1997/12/24 05:19:26 itojun Exp $
 */

#include <math.h>
#include "mgp.h"

GC gcmask;
GC gcall;
GC gcnone;
GC gcfore;
GC gcback;
GC gcfb;
GC gcor;
GC gcpen;
GC gcgreen;
GC gcyellow;
GC gcred;
GC gc_pl;
GC gc_plrev;
GC gc_pta;
GC gc_ptk;

static XWindowAttributes xa;

u_int window_x;
u_int window_y;

void
init_win1()
{
	if ((display = XOpenDisplay(NULL)) == NULL) {
		fprintf(stderr, "Can't open display\n");
		exit(-1);
	}

	XGetWindowAttributes(display, DefaultRootWindow(display), &xa);
	if (window_width == 0 || window_height == 0) {
		window_width = xa.width;
		window_height = xa.height;
	}
	if (mgp_flag & FL_OVER) {
		/* reduce the size slightly to enable window operations */
		window_width  = (int)(window_width * 96 / 100);
		window_height = (int)(window_height * 96 / 100);
	}
	if (window_x < 0)
		window_x = xa.width + window_x;

	if (window_y < 0)
		window_y = xa.height + window_y;

	char_size = window_height / 10;
	(void)get_color(DEFAULT_FORE, &fore_color);
	ctrl_color = fore_color;

	if (!strlen(back_clname))
		strcpy(back_clname, DEFAULT_BACK);
	(void)get_color(back_clname, &back_color);
}

void
init_win2()
{
	XSetWindowAttributes xsa;
	u_long mask = 0;
	u_int i;
	u_long color;

	screen = DefaultScreen(display);
	depth = DefaultDepth(display, screen);
	depth_mask = 1;
	for (i = depth; i; i--)
		depth_mask *= 2;
	depth_mask--;
	/* depth_mask is the max value of pixel */

	window = XCreateSimpleWindow(display, DefaultRootWindow(display),
		0, 0, window_width, window_height, 0, fore_color, back_color);

	pixmap  = XCreatePixmap(display, window, xa.width, xa.height, depth);
	maskpix = XCreatePixmap(display, window, xa.width, xa.height, depth);

	gcall = XCreateGC(display, window, 0, 0);
	gcnone = XCreateGC(display, window, 0, 0);
	gcfore = XCreateGC(display, window, 0, 0);
	gcback = XCreateGC(display, window, 0, 0);
	gcfb = XCreateGC(display, window, 0, 0);
	gcmask = XCreateGC(display, window, 0, 0);
	gcor = XCreateGC(display, window, 0, 0);
	gcpen = XCreateGC(display, window, 0, 0);
	gcgreen = XCreateGC(display, window, 0, 0);
	gcyellow = XCreateGC(display, window, 0, 0);
	gcred = XCreateGC(display, window, 0, 0);

	XSetFunction(display, gcall, GXcopy);
	XSetForeground(display, gcall, depth_mask);
	XSetBackground(display, gcall, 0);
	XSetFunction(display, gcnone, GXcopy);
	XSetForeground(display, gcnone, 0);
	XSetBackground(display, gcnone, 0);
	XSetFunction(display, gcmask, GXandInverted);
	XSetForeground(display, gcmask, depth_mask);
	XSetBackground(display, gcmask, 0);
	XSetFunction(display, gcfore, GXcopy);
	XSetForeground(display, gcfore, fore_color);
	XSetBackground(display, gcfore, 0);
	XSetFunction(display, gcback, GXcopy);
	XSetForeground(display, gcback, back_color);
	XSetFunction(display, gcfb, GXcopy);
	XSetForeground(display, gcfb, fore_color);
	XSetBackground(display, gcfb, back_color);
	XSetFunction(display, gcor, GXor);
	XSetForeground(display, gcor, depth_mask);
	XSetBackground(display, gcor, 0);
	(void)get_color("red", &color);
	XSetForeground(display, gcpen, color);
	XSetForeground(display, gcred, color);
	(void)get_color("green", &color);
	XSetForeground(display, gcgreen, color);
	(void)get_color("yellow", &color);
	XSetForeground(display, gcyellow, color);

	if (!(mgp_flag & FL_OVER)) {
		xsa.override_redirect = True;
		mask |= CWOverrideRedirect;
		XSetIconName(display, window, mgp_fname);
	}
	xsa.backing_store = Always;
	mask |= CWBackingStore;

	XChangeWindowAttributes(display, window, mask, &xsa);

	XSelectInput(display, window, EVENT_DEFAULT);
	XMapRaised(display, window);

	if (!(mgp_flag & FL_OVER))
		XMoveWindow(display, window, window_x, window_y);
	XFlush(display);
	if (!(mgp_flag & FL_OVER)) {
		Window dummy;
		int revert;
		XGetInputFocus(display, &dummy, &revert);
		XSetInputFocus(display, window, revert, CurrentTime);
	}

	gc_pl = XCreateGC(display, window, 0, 0);
	gc_plrev = XCreateGC(display, window, 0, 0);
	gc_pta = XCreateGC(display, window, 0, 0);
	gc_ptk = XCreateGC(display, window, 0, 0);
	plfs = XLoadQueryFont(display, PAGELIST_FONT);
	plkfs = XLoadQueryFont(display, PAGELIST_KFONT);
	XSetFont(display, gc_pl, plfs->fid);
	XSetFont(display, gc_plrev, plfs->fid);
	XSetFont(display, gc_pta, plfs->fid);
	XSetFont(display, gc_ptk, plkfs->fid);
	XSetFunction(display, gc_pl, GXcopy);
	XSetFunction(display, gc_plrev, GXcopy);
	XSetFunction(display, gc_pta, GXcopy);
	XSetFunction(display, gc_ptk, GXcopy);
	XSetForeground(display, gc_pl, BlackPixel(display, screen));
	XSetBackground(display, gc_pl, WhitePixel(display, screen));
	XSetForeground(display, gc_plrev, WhitePixel(display, screen));
	XSetBackground(display, gc_plrev, BlackPixel(display, screen));
	pl_fh = (plfs->max_bounds.ascent + plfs->max_bounds.descent) * 1.2;
	pl_fw = plfs->max_bounds.rbearing + plfs->max_bounds.lbearing;
	pl_nx = (window_width - 20) / pl_fw / 3;   /* list pages in x axis */
	pl_ny = maxpage / pl_nx + 1;    /* list pages in y axis */
}

int
get_color(colorname, value)
	u_char *colorname;
	u_long *value;
{
	Colormap cmap;
	XColor c0, c1;

	cmap = DefaultColormap(display, screen);
	if (XAllocNamedColor(display, cmap, colorname, &c1, &c0) == 0)
		return -1;
	if (value)
		*value = c1.pixel;
	return 0;
}
